# Ugly hack to allow import from the root folder
import sys
import os
sys.path.insert(0, os.path.abspath('..'))

import unittest
from boba.constraintparser import ConstraintParser, ParseError
from boba.conditionparser import ConditionParser, TokenType
from boba.parser import Parser


def abs_path(rel_path):
    return os.path.join(os.path.dirname(__file__), rel_path)


def read_wrapper(spec, ps):
    ConstraintParser(spec).read_constraints(ps.code_parser, ps.dec_parser)


class TestConstraintParser(unittest.TestCase):

    def test_read_json(self):
        base = abs_path('./specs/')
        ps = Parser(base+'script3-1.py')
        cp = ConstraintParser(ps.spec)
        cs = cp.read_constraints(ps.code_parser, ps.dec_parser)
        self.assertEqual(len(cs), 2)

    def test_link(self):
        base = abs_path('./specs/')
        ps = Parser(base + 'script3-7.py')
        cp = ConstraintParser(ps.spec)
        cs = cp.read_constraints(ps.code_parser, ps.dec_parser)
        self.assertEqual(len(cs), 10)

    def test_condition_parser(self):
        cond = ''
        ConditionParser(cond).parse()

        cond = 'a == b'
        _, decs = ConditionParser(cond).parse()
        self.assertListEqual(['a', 'b'], [d.value for d in decs])

        cond = 'a.index == 1'
        _, decs = ConditionParser(cond).parse()
        self.assertListEqual(['a', '1'], [d.value for d in decs])
        self.assertListEqual([TokenType.index_var, TokenType.number],
                             [d.type for d in decs])

        cond = 'a = 2.5'
        _, decs = ConditionParser(cond).parse()
        self.assertListEqual(['a', '2.5'], [d.value for d in decs])
        self.assertListEqual([TokenType.var, TokenType.number],
                             [d.type for d in decs])

        cond = 'a.index == b.index'  # .index not allowed on RHS, should fail
        with self.assertRaises(ParseError):
            ConditionParser(cond).parse()

        cond = '1 2 a b 4'  # we did not check other semantics ...
        ConditionParser(cond).parse()

    def test_eval(self):
        """ Evaluation of various conditions """
        # expr and expr
        base = abs_path('./specs/')
        ps = Parser(base + 'script3-6.py', base)
        ps.main(verbose=False)
        self.assertEqual(ps.wrangler.counter, 2)

        # expr or expr
        ps.spec['constraints'] = [{"block": "D", "condition": "a == if or B == b1"}]
        ps._parse_constraints()
        ps.main(verbose=False)
        self.assertEqual(ps.wrangler.counter, 6)

        # expr and (expr or expr)
        ps.spec['constraints'] = [{"block": "D", "condition": "a == if and (B == b1 or B == b2)"}]
        ps._parse_constraints()
        ps.main(verbose=False)
        self.assertEqual(ps.wrangler.counter, 4)

        # testing !=
        ps.spec['constraints'] = [{"block": "D", "condition": "a != if"}]
        ps._parse_constraints()
        ps.main(verbose=False)
        self.assertEqual(ps.wrangler.counter, 4)

        # testing >=
        ps.spec['constraints'] = [{"block": "D", "condition": "a.index >= 1"}]
        ps._parse_constraints()
        ps.main(verbose=False)
        self.assertEqual(ps.wrangler.counter, 4)

        # testing index
        ps.spec['constraints'] = [{"block": "D", "condition": "b.index == 1"}]
        ps._parse_constraints()
        ps.main(verbose=False)
        self.assertEqual(ps.wrangler.counter, 4)

        # testing option with integer type
        ps.spec['constraints'] = [{"block": "D", "condition": "b == 0"}]
        ps._parse_constraints()
        ps.main(verbose=False)
        self.assertEqual(ps.wrangler.counter, 4)

        # testing option with float type
        ps.spec['constraints'] = [{"block": "D", "condition": "b == 1.5"}]
        ps._parse_constraints()
        ps.main(verbose=False)
        self.assertEqual(ps.wrangler.counter, 4)

        # testing unmade decision
        ps.spec['constraints'] = [{"block": "A", "condition": "b.index == 0"}]
        ps._parse_constraints()
        ps.main(verbose=False)
        self.assertEqual(ps.wrangler.counter, 0)

        # testing if the decision is made when the block depends on a variable
        # inside the block
        ps.spec['constraints'] = [{"block": "B", "condition": "b.index == 0"}]
        ps._parse_constraints()
        ps.main(verbose=False)
        self.assertEqual(ps.wrangler.counter, 0)

    def test_condition_syntax(self):
        """ Does the condition code contain python syntax error? """

        base = abs_path('./specs/')
        ps = Parser(base+'script3-1.py', base)

        spec = {'constraints': [{'block': 'A', 'condition': 'B=b1'}]}
        with self.assertRaises(ParseError):
            read_wrapper(spec, ps)

        spec = {'constraints': [{'block': 'A', 'condition': 'B b1'}]}
        with self.assertRaises(ParseError):
            read_wrapper(spec, ps)

        spec = {'constraints': [{'block': 'A', 'condition': 'B == 2.5'}]}
        read_wrapper(spec, ps)

    def test_json_syntax(self):
        """ Test various possibilities to specify constraints in JSON """

        base = abs_path('./specs/')
        ps = Parser(base+'script3-1.py', base)

        # empty - should parse
        spec = {}
        read_wrapper(spec, ps)

        # empty array - should parse
        spec = {'constraints': []}
        read_wrapper(spec, ps)

        # empty element - should fail
        spec = {'constraints': [{}]}
        with self.assertRaises(ParseError):
            read_wrapper(spec, ps)

        # no matching block - should fail
        spec = {'constraints': [{'block': 'a'}]}
        with self.assertRaises(ParseError):
            read_wrapper(spec, ps)

        # no matching variable - should fail
        spec = {'constraints': [{'variable': 'c'}]}
        with self.assertRaises(ParseError):
            read_wrapper(spec, ps)

        # loner option - should fail
        spec = {'constraints': [{'option': 'a1'}]}
        with self.assertRaises(ParseError):
            read_wrapper(spec, ps)

        # loner block - should parse
        spec = {'constraints': [{'block': 'A', 'condition': 'B==b1'}]}
        read_wrapper(spec, ps)

        # block and option - should parse
        spec = {'constraints': [{'block': 'A', 'option': 'a1', 'condition': 'B==b1'}]}
        read_wrapper(spec, ps)

        # variable and option - should parse
        spec = {'constraints': [{'variable': 'a', 'option': '2.5', 'condition': 'B==b1'}]}
        read_wrapper(spec, ps)

        # weird option - should parse
        # fixme: {'option': '[1,2]'} will fail
        spec = {'constraints': [{'variable': 'c', 'option': '[1, 2]', 'condition': 'B==b1'}]}
        read_wrapper(spec, ps)

        # variables in condition do not match - should fail
        spec = {'constraints': [{'block': 'A', 'condition': 'H==b1'}]}
        with self.assertRaises(ParseError):
            read_wrapper(spec, ps)

        # variables in condition do not match - should fail
        spec = {'constraints': [{'block': 'A', 'condition': 'H.index==1'}]}
        with self.assertRaises(ParseError):
            read_wrapper(spec, ps)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

import os
import re
import pandas as pd
import boba.util as util

STR_MAX = 1024


class CSVMerger:
    def __init__(self, pattern, base, out, delimiter=','):
        self.pattern = pattern
        self.base = base
        self.out = out
        self.delimiter = delimiter

    def _fn_func(self, i):
        return self.pattern.format(i)

    def _to_regex(self):
        """ Convert the string pattern to regex. """
        i = self.pattern.find('{}')
        if i < 0:
            util.print_fail('Invalid pattern: {}'.format(self.pattern))
            exit(1)

        rg = re.compile('^' + re.escape(self.pattern[:i]) + '(\d+)' +
                        re.escape(self.pattern[i+2:]))
        return rg

    def get_files(self):
        """ Get a list of universe indices in the folder that matches given
        pattern. The indices are sorted."""
        idx = []
        for f in os.listdir(self.base):
            m = re.match(self._to_regex(), f)
            if m:
                idx.append(int(m.group(1)))
        idx.sort()
        return idx

    def merge(self):
        """ Merge the CSV files into one file """
        result = pd.DataFrame()
        for i in self.get_files():
            # read the file
            df = pd.read_csv(os.path.join(self.base, self._fn_func(i)),
                             delimiter=self.delimiter,
                             converters={i: str for i in range(0, STR_MAX)})
            n = len(list(df.columns))

            # augment
            df['uid'] = i

            # rearrange columns
            cols = list(df.columns)
            cols = cols[n:] + cols[:n]
            df = df[cols]

            # merge with previous results
            result = pd.concat([result, df], axis=0, sort=False)

        return result

    def main(self):
        res = self.merge()
        res.to_csv(self.out, index=False)



#!/usr/bin/env python3

# Ugly hack to allow import from the root folder
import sys
import os
sys.path.insert(0, os.path.abspath('..'))

import unittest
from boba.blocksyntaxparser import BlockSyntaxParser, ParseError


class TestBlockParser(unittest.TestCase):

    def test_steps(self):
        line = '# --- (A) remove_outlier'
        self.assertTrue(BlockSyntaxParser.can_parse(line))
        bp = BlockSyntaxParser(line)
        self.assertEqual(bp.i, 0)
        bp._read_next()
        self.assertEqual(bp.i, 5)
        bp._read_next()
        self.assertEqual(bp.i, 9)
        bp._read_next()
        self.assertEqual(bp.i, len(line))
        bp._read_next()
        self.assertEqual(bp.i, len(line))

    def test_can_parse(self):
        self.assertTrue(BlockSyntaxParser.can_parse('   # --- comment'))
        self.assertTrue(BlockSyntaxParser.can_parse('   # ---comment   '))
        self.assertFalse(BlockSyntaxParser.can_parse('#--- comment'))
        self.assertFalse(BlockSyntaxParser.can_parse('# --'))

    def test_syntax(self):
        line = '# --- (A) remove_outlier'
        self.assertTrue(BlockSyntaxParser.can_parse(line))
        bid, par, opt, cond = BlockSyntaxParser(line).parse()
        self.assertEqual(bid, 'A:remove_outlier')
        self.assertEqual(par, 'A')
        self.assertEqual(opt, 'remove_outlier')

        line = '# --- (A) remove outlier'
        with self.assertRaises(ParseError):
            BlockSyntaxParser(line).parse()

        line = '# --- ((A)) name'
        with self.assertRaises(ParseError):
            BlockSyntaxParser(line).parse()

        line = '# --- ( A)'
        bid, par, opt, cond = BlockSyntaxParser(line).parse()
        self.assertEqual(bid, 'A')
        self.assertEqual(par, '')
        self.assertEqual(opt, '')

    def test_condition(self):
        line = '# --- (A) a1 @if B == b1'
        bid, par, opt, cond = BlockSyntaxParser(line).parse()
        self.assertEqual(bid, 'A:a1')
        self.assertEqual(par, 'A')
        self.assertEqual(opt, 'a1')
        self.assertEqual(cond['block'], 'A')
        self.assertEqual(cond['option'], 'a1')
        self.assertEqual(cond['condition'], 'B == b1')

        line = '# --- (A) @if B == b1'
        bid, par, opt, cond = BlockSyntaxParser(line).parse()
        self.assertEqual(cond['block'], 'A')
        self.assertEqual(cond['condition'], 'B == b1')
        self.assertNotIn('option', cond)

        line = '# --- (A) remove outlier @if B == b1'
        with self.assertRaises(ParseError):
            BlockSyntaxParser(line).parse()

    def test_whitespace(self):
        line = '\t\t# --- (A) name'
        bid, par, opt, cond = BlockSyntaxParser(line).parse()
        self.assertEqual(par, 'A')
        self.assertEqual(opt, 'name')

        line = '    # --- (A) name    \t'
        bid, par, opt, cond = BlockSyntaxParser(line).parse()
        self.assertEqual(par, 'A')
        self.assertEqual(opt, 'name')

        line = '# ---(A)socrowded'
        bid, par, opt, cond = BlockSyntaxParser(line).parse()
        self.assertEqual(par, 'A')
        self.assertEqual(opt, 'socrowded')

    def test_id_syntax(self):
        line = '# --- (C1)'
        bid, par, opt, cond = BlockSyntaxParser(line).parse()
        self.assertEqual(bid, 'C1')

        line = '# --- (aXa) '
        bid, par, opt, cond = BlockSyntaxParser(line).parse()
        self.assertEqual(bid, 'aXa')

        line = '# --- (my_variable) \t'
        bid, par, opt, cond = BlockSyntaxParser(line).parse()
        self.assertEqual(bid, 'my_variable')

        # ID must start with a letter
        line = '# --- (12)'
        with self.assertRaisesRegex(ParseError, '(?i)invalid identifier'):
            BlockSyntaxParser(line).parse()

        line = '# --- (_start)'
        with self.assertRaisesRegex(ParseError, '(?i)invalid identifier'):
            BlockSyntaxParser(line).parse()


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

import json
from dataclasses import dataclass
from .baseparser import ParseError, BaseParser


@dataclass
class Decision:
    var: str
    value: list
    desc: str = ''


class DecisionParser(BaseParser):
    def __init__(self):
        super(DecisionParser, self).__init__('')
        self.decisions = {}

    @staticmethod
    def _is_syntax_start(ch):
        return ch == '{'

    @staticmethod
    def _is_id_token(s):
        if not BaseParser._is_id_start(s[0]):
            return False

        for ch in s:
            if not BaseParser._is_id(ch):
                return False

        return True

    @staticmethod
    def _read_value(s):
        try:
            res = list(s)
        except ValueError:
            raise ParseError('Cannot handle value "{}"'.format(s))

        if len(s) == 0:
            raise ParseError('Cannot handle decision value "[]"')
        return res

    def _read_json_safe(self, obj, field):
        if field not in obj:
            raise ParseError('Cannot find "{}" in json'.format(field))
        return obj[field]

    def _check_type(self, var, fun, msg):
        if not fun(var):
            raise ParseError('Cannot handle {} "{}"'.format(msg, var))
        return var

    def verify_naming(self, reserved):
        """
        Verify if the decision names collide with any other variable names.
        :param reserved: A list of reserved names.
        :return:
        """
        for w in reserved:
            if w in self.decisions:
                raise ParseError('Duplicate variable/block name "{}"'.format(w))

    def read_decisions(self, spec):
        """
        Read decisions from the JSON spec.
        :return: a dict of decisions
        """
        dec_spec = spec['decisions'] if 'decisions' in spec else []
        for d in dec_spec:
            desc = d['desc'] if 'desc' in d else 'Decision {}'.format(d['var'])

            var = self._check_type(self._read_json_safe(d, 'var'),
                                   DecisionParser._is_id_token, 'id')
            value = self._read_value(self._read_json_safe(d, 'options'))

            # check if two variables have the same name
            if var in self.decisions:
                raise ParseError('Duplicate variable name "{}"'.format(var))

            decision = Decision(var, value, desc)
            self.decisions[var] = decision

        return self.decisions

    def get_num_alt(self, dec):
        """
        Return the number of possible alternatives.
        For discrete type, it's the number of options.
        :param dec: variable ID of a decision
        :return: number
        """
        return len(self.decisions[dec].value)

    def get_alt(self, dec, i_alt):
        """
        Return the i-th alternative. For discrete type, it's simply the i-th
        option specified.
        :param dec: variable ID of a decision
        :param i_alt: which alternative
        :return: value of the option
        """
        return self.decisions[dec].value[i_alt]

    def get_cross_prod(self):
        """
        Get the maximum possible cardinality of options, computed as a cross
        product of all decisions.
        :return: number
        """
        ret = 1
        for dec in self.decisions:
            ret *= self.get_num_alt(dec)
        return ret

    def get_decs(self):
        """Get a list of decision names."""
        return [i for i in self.decisions.keys()]

    def gen_code(self, template, dec_id, i_alt):
        """
        Replace the placeholder variable in a template chunk.
        :param template: a chunk of code with only one placeholder
        :param dec_id: variable ID of the decision
        :param i_alt: which alternative
        :return: {string, string} replaced code and the value at this parameter
        """
        v = self.get_alt(dec_id, i_alt)

        # assuming the placeholder var is always at the end
        # which is true given how we chop up the chunks
        return template + str(v), str(v)

    def parse_code(self, line):
        """
        Find placeholder variables in a line of code.
        :param line: a line of code
        :return: (variables, chunks) a list of found variables and a list of
                 code chunks, each containing one variable
        """
        code = []
        res = []
        j = 0
        i_start = 0
        self.i = 0
        self.line = line

        while not self._is_end():
            if self._is_syntax_start(self._peek_char()):
                token = self._read_while(lambda ch: ch == '{')
                if len(token) < 2:
                    continue
                i_start = self.i - 2

                # read variable identifier
                if not self._is_id_start(self._peek_char()):
                    continue
                val = self._read_while(self._is_id)
                if len(val) == 0:
                    continue

                # read definition, if any
                df = None
                self._read_while(self._is_whitespace)
                if not self._is_end() and self._peek_char() == '=':
                    self._next_char()
                    # problem: the variable value can't contain "}"
                    df = self._read_while(lambda ch: ch != '}')
                    self._read_while(self._is_whitespace)

                token = self._read_while(lambda ch: ch == '}', max_len=2)
                if len(token) < 2:
                    continue

                # read succeeds
                code.append(line[j:i_start])
                res.append(val)
                j = self.i

                # parse and save definition
                if df:
                    try:
                        df = json.loads('[{}]'.format(df))
                        decision = Decision(val, df, '')
                        if val in self.decisions:
                            msg = 'Duplicate variable definition "{}"'
                            raise ParseError(msg.format(val))
                        self.decisions[val] = decision
                    except ValueError:
                        msg = 'Cannot parse variable definition:\n{}'
                        raise ParseError(msg.format(df))
            else:
                self._next_char()

        code.append(line[j:])
        return res, code



# -*- coding: utf-8 -*-

import json
import os
from textwrap import wrap
from dataclasses import dataclass, field
from typing import List

from .baseparser import ParseError
from .codeparser import CodeParser
from .graphparser import GraphParser
from .graphanalyzer import GraphAnalyzer, InvalidGraphError
from .decisionparser import DecisionParser
from .constraintparser import ConstraintParser
from .lang import LangError, Lang
from .wrangler import Wrangler
from .adg import ADG
import boba.util as util


@dataclass
class History:
    """
    A class for keeping track of the choices made in each universe.

    path: index of the code path.
    filename: file name of the universe.
    decisions: placeholder variables and the options they took.
    skipped: nodes that are skipped.
    """
    path: int
    filename: str = ''
    decisions: List = field(default_factory=lambda: [])
    skipped: List = field(default_factory=lambda: [])


@dataclass
class DecRecord:
    """ A class for what options a parameter took. """
    parameter: str = ''
    option: str = ''
    idx: int = -1


class Parser:

    """ Parse everything """

    def __init__(self, f1, out='.', lang=''):
        self.fn_script = f1
        self.out = os.path.join(out, 'multiverse/')

        self.paths = []
        self.history = []
        self.constraints = {}

        # init parser class
        self.code_parser = CodeParser()
        self.dec_parser = DecisionParser()
        self.adg = ADG()

        # parse
        self._parse_blocks()
        self.spec = self.code_parser.spec
        self._parse_decs()
        self._parse_graph()
        self._parse_constraints()

        # init helper class
        try:
            self.lang = Lang(lang, f1)
            self.wrangler = Wrangler(self.spec, self.lang, self.out)
        except LangError as e:
            self._throw(e.args[0])
        except ParseError as e:
            self._throw_spec_error(e.args[0])

    def _throw(self, msg):
        util.print_fail(msg)
        exit(1)

    def _indent(self, msg):
        return '\n'.join(['\t' + l for l in msg.split('\n')])

    @staticmethod
    def _rename_dec(dec):
        return '_' + dec

    def _throw_parse_error(self, msg):
        prefix = 'In parsing file "{}":\n'.format(self.fn_script)
        self._throw(prefix + self._indent(msg))

    def _throw_spec_error(self, msg):
        prefix = 'In parsing boba config:\n' + json.dumps(self.spec) + '\n'
        self._throw(prefix + self._indent(msg))

    def _parse_blocks(self):
        """ Make a pass over the template, parsing block declarations and
        placeholder variables inside the code."""
        with open(self.fn_script, 'r') as f:
            try:
                self.code_parser.parse(self.dec_parser, f)
            except ParseError as e:
                self._throw_parse_error(e.args[0])

    def _parse_decs(self):
        """ Parse decisions in the spec. """
        try:
            self.dec_parser.read_decisions(self.spec)
        except ParseError as e:
            self._throw_spec_error(e.args[0])

        # check if used variables has been declared
        for v in self.code_parser.used_vars:
            if v not in set(self.dec_parser.decisions.keys()):
                msg = 'Cannot find matching variable "{}" in spec'.format(v)
                self._throw_spec_error(msg)

    def _match_nodes(self, nodes):
        """ Nodes in spec and script should match. """
        blocks = self.code_parser.get_block_names()

        for nd in nodes:
            if nd not in blocks:
                self._throw_spec_error('Cannot find matching node "{}" in script'.format(nd))

        for nd in blocks:
            # ignore special nodes inserted by us
            if not nd.startswith('_') and nd not in nodes:
                util.print_warn('Cannot find matching node "{}" in graph spec'.format(nd))

    def _parse_graph(self):
        graph_spec = self.spec['graph'] if 'graph' in self.spec else []

        try:
            gp = GraphParser(graph_spec)
            nodes, edges = gp.parse()

            # if the user didn't specify the graph, use default
            if len(nodes) == 0:
                nodes, edges = gp.create_default_graph(self.code_parser.order)

            # check if graph nodes matches block IDs, and vice versa
            self._match_nodes(nodes)
            # check if any name of nodes collides with variable names
            self.dec_parser.verify_naming(nodes)
            # expand the graph with options
            nodes, edges = gp.replace_graph(self.code_parser.get_decisions())

            # analyze the graph to get paths
            self.paths = GraphAnalyzer(nodes, edges).analyze()
            # an ugly way to handle the artificial _start node
            if '_start' in self.code_parser.blocks and '_start' not in nodes:
                for p in self.paths:
                    p.insert(0, '_start')
                if len(self.paths) == 0:
                    self.paths = [['_start']]

            # save data to adg
            self.adg.set_graph(nodes, edges)
        except ParseError as e:
            self._throw_spec_error(e.args[0])
        except InvalidGraphError as e:
            self._throw_spec_error(e.args[0])

    def _parse_constraints(self):
        try:
            cp = ConstraintParser(self.spec)
            self.constraints = cp\
                .read_constraints(self.code_parser, self.dec_parser)
            # save intermediate data for ADG
            self.adg.set_constraints(cp.links, cp.procedural)
        except ParseError as e:
            self._throw_spec_error(e.args[0])

    def _eval_constraint(self, history, con):
        """ See if the constraint holds true given the choices made. """
        con = self.constraints[con].condition
        paths, bdecs = self._nice_path(self._get_skipped_path(history))

        # A dict where the key is each parameter and the value is the chosen
        # option. For ordinary blocks, key and value are the same.
        res = {}
        for p in paths:
            res[p] = p
        for d in bdecs:
            # note that block parameter will override with the actual option
            res[d.parameter] = d.option
        for d in self.dec_parser.get_decs():
            # unmade decisions will have value None and index -1
            res[d] = None
            res[ConstraintParser.make_index_var(d)] = -1
        for d in history.decisions:
            res[d.parameter] = d.option
            res[ConstraintParser.make_index_var(d.parameter)] = d.idx

        # now evaluate
        return eval(con, res)

    def _get_code_paths(self):
        """ Convert paths of block to paths of code chunk """

        res = []

        for path in self.paths:
            pt = []
            for nd in path:
                # replace the node by its chunks
                chunks = [(nd, ch) for ch in self.code_parser.blocks[nd].chunks]
                pt.extend(chunks)
            res.append(pt)

        return res

    def _code_gen_recur(self, path, i, code, history):
        """
        Generate code recursively.

        :param path: the code path we're on.
        :param i: the index of the current node in the code path.
        :param code: the generated code so far.
        :param history: record the choices made.
        """

        if i >= len(path):
            # write file
            fn = self.wrangler.write_universe(code)

            # record history
            history.filename = fn
            self.history.append(history)
        else:
            nd, chunk = path[i]

            # check if the block has constraints attached to it
            names = [nd, nd.split(':')[0]]
            for n in names:
                if n in self.constraints and\
                        not self._eval_constraint(history, n):
                    # constraint met
                    if self.constraints[n].skip:
                        # skip the node and continue
                        history.skipped.append(nd)
                        self._code_gen_recur(path, i + 1, code, history)
                        return
                    else:
                        # abort codegen
                        return

            if chunk.variable != '':
                # check if we have already encountered the placeholder variable
                prev_idx = None
                for d in history.decisions:
                    if d.parameter == chunk.variable:
                        prev_idx = d.idx

                if prev_idx is not None:
                    # use the previous value
                    snippet, opt = self.dec_parser.gen_code(chunk.code, chunk.variable, prev_idx)
                    self._code_gen_recur(path, i+1, code+snippet, history)
                else:
                    # expand the decision
                    num_alt = self.dec_parser.get_num_alt(chunk.variable)
                    for k in range(num_alt):
                        # check if the option has constraints attached to it
                        # always check by index, rather than actual value
                        v = ConstraintParser.make_index_var(chunk.variable)
                        v = '{}:{}'.format(v, k)
                        if v in self.constraints and \
                                not self._eval_constraint(history, v):
                            # constraint met, abort
                            continue

                        # code gen
                        snippet, opt = self.dec_parser.gen_code(chunk.code, chunk.variable, k)
                        decs = [a for a in history.decisions]
                        decs.append(DecRecord(chunk.variable, opt, k))
                        self._code_gen_recur(path, i+1, code + snippet,
                                             History(history.path, '', decs))
            else:
                code += chunk.code
                self._code_gen_recur(path, i+1, code, history)

    def _code_gen(self):
        paths = self._get_code_paths()

        self.wrangler.counter = 0  # keep track of file name
        self.history = []          # keep track of choices made for each file
        self.wrangler.col = 2 + len(self.dec_parser.get_decs())\
            + len(self.code_parser.get_decisions())

        self.wrangler.create_dir()
        for idx, p in enumerate(paths):
            self._code_gen_recur(p, 0, '', History(idx))

        # output a script to execute all universes
        self.wrangler.write_sh()

    @staticmethod
    def _nice_path(path):
        """ Convert the path containing block options back to the simpler path
         containing only block parameters."""
        ps = [p.split(':')[0] for p in path]
        decs = [DecRecord(p.split(':')[0], p.split(':')[1]) for p in path if ':' in p]
        return ps, decs

    def _get_skipped_path(self, h):
        """ Get the history's path, with skipped nodes removed. """
        sk = set(h.skipped)
        return [nd for nd in self.paths[h.path] if nd not in sk]

    def _write_csv(self):
        rows = []
        decs = self.dec_parser.get_decs() +\
            [b for b in self.code_parser.get_decisions()]
        ops = self.wrangler.get_outputs()
        rows.append(['Filename', 'Code Path'] + decs + ops)
        for h in self.history:
            paths, bdecs = self._nice_path(self._get_skipped_path(h))
            row = [h.filename, '->'.join(paths)]
            mp = {}
            for d in h.decisions:
                mp[d.parameter] = d.option
            for d in bdecs:
                mp[d.parameter] = d.option
            for d in decs:
                value = mp[d] if d in mp else ''
                row.append(value)
            rows.append(row)
        self.wrangler.write_summary(rows)

    def _write_server_config(self):
        self.adg.create(self.code_parser.blocks)
        res = self.adg.output()

        # get the options of decision blocks and placeholders
        lookup = self.code_parser.get_decisions()
        for d in lookup:
            lookup[d] = [v.split(':')[1] for v in lookup[d]]
        decs = self.dec_parser.decisions
        for d in decs:
            lookup[d] = decs[d].value

        # now wrangle the decisions
        decs = self.adg.get_used_decs()
        decs = [{"var": d, "options": lookup[d]} for d in decs]
        res["decisions"] = decs

        # save
        self.wrangler.write_overview_json(res)

    def _print_summary(self):
        w = 80
        max_rows = 10

        print('=' * w)
        print('{:<20}{:<30}{:<30}'.format('Filename', 'Code Path', 'Decisions'))
        print('=' * w)
        for idx, h in enumerate(self.history):
            paths, bdecs = self._nice_path(self._get_skipped_path(h))
            path = wrap('->'.join(paths), width=27)
            decs = ['{}={}'.format(d.parameter, d.option) for d in bdecs + h.decisions]
            decs = wrap(', '.join(decs), width=30)
            max_len = max(len(decs), len(path))

            for r in range(max_len):
                c1 = h.filename if r == 0 else ''
                c2 = path[r] if r < len(path) else ''
                c3 = decs[r] if r < len(decs) else ''

                print('{:<20}'.format(c1), end='')
                print('{:<30}'.format(c2), end='')
                print('{:<30}'.format(c3))
            print('-' * w)

            if idx >= max_rows - 1:
                print('... {} more rows'.format(len(self.history) - max_rows))
                break

    def _warn_size(self):
        cap = self.dec_parser.get_cross_prod() * len(self.paths)
        if cap > 1024:
            rs = input('\nBoba may create as many as {} scripts. '
                       'Proceed (y/n)?\n'.format(cap))
            if not rs.strip().lower().startswith('y'):
                print('Aborted.')
                exit(0)

    def main(self, verbose=True):
        self._warn_size()
        self._code_gen()
        self._write_csv()
        self._write_server_config()
        if verbose:
            self._print_summary()



#!/usr/bin/env python3

# Ugly hack to allow import from the root folder
import sys
import os
sys.path.insert(0, os.path.abspath('..'))

import unittest
from boba.graphparser import GraphParser, Edge, ParseError


class TestParser(unittest.TestCase):
    def test_good_specs(self):
        spec = ['A -> B -> C1', 'B->C2']
        nodes, edges = GraphParser(spec).parse()
        self.assertSetEqual(nodes, {'A', 'B', 'C1', 'C2'})
        exp_edges = {Edge('A', 'B'), Edge('B', 'C1'), Edge('B', 'C2')}
        self.assertSetEqual(edges, exp_edges)

    def test_weird_specs(self):
        spec = ['a->a->a->a  b ']
        nds, eds = GraphParser(spec).parse()
        self.assertSetEqual(nds, {'a', 'b'})
        self.assertSetEqual(eds, {Edge('a', 'a')})

        spec = ['a  b', 'c']
        nds, eds = GraphParser(spec).parse()
        self.assertSetEqual(nds, {'a', 'b', 'c'})
        self.assertSetEqual(eds, set())

        spec = ['a->b c->b']
        nds, eds = GraphParser(spec).parse()
        self.assertSetEqual(nds, {'a', 'b', 'c'})
        self.assertSetEqual(eds, {Edge('a', 'b'), Edge('c', 'b')})

    def test_syntax_error(self):
        spec = ['my_first_node -> my_second_node']
        nds, eds = GraphParser(spec).parse()
        self.assertSetEqual(nds, {'my_first_node', 'my_second_node'})
        self.assertSetEqual(eds, {Edge('my_first_node', 'my_second_node')})

        spec = ['_start -> _end']
        with self.assertRaisesRegex(ParseError, '(?i)cannot handle character'):
            GraphParser(spec).parse()

        spec = ['-> B']
        with self.assertRaisesRegex(ParseError, '(?i)source node'):
            GraphParser(spec).parse()

        spec = ['A -> B ->']
        with self.assertRaisesRegex(ParseError, '(?i)target node'):
            GraphParser(spec).parse()

        spec = ['A - B']
        with self.assertRaises(ParseError):
            GraphParser(spec).parse()

        spec = ['A->B->C, B->D']
        with self.assertRaises(ParseError):
            GraphParser(spec).parse()


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

"""Console script."""
import click
import os
import subprocess
from .parser import Parser
from .output.csvmerger import CSVMerger


@click.command()
@click.option('--script', '-s', help='Path to template script',
              default='./template.py', show_default=True)
@click.option('--out', help='Output directory',
              default='.', show_default=True)
@click.option('--lang', help='Language, can be python/R [default: inferred from file extension]',
              default='')
def compile(script, out, lang):
    """Generate multiverse analysis from specifications."""

    check_path(script)

    click.echo('Creating multiverse from {}'.format(script))
    ps = Parser(script, out, lang)
    ps.main()

    ex = """To execute the multiverse, run the following commands:
    boba run --all
    """.format(os.path.join(out, 'multiverse'))
    click.secho('Success!', fg='green')
    click.secho(ex, fg='green')


def check_path(p):
    """Check if the path exists"""
    if not os.path.exists(p):
        msg = 'Error: Path "{}" does not exist.'.format(p)
        print_help(msg)


def print_help(err=''):
    """Show help message and exit."""
    ctx = click.get_current_context()
    click.echo(ctx.get_help())

    if err:
        click.echo('\n' + err)
    ctx.exit()


@click.command()
@click.argument('num', nargs=1, default=-1)
@click.option('--all', '-a', 'run_all', is_flag=True,
              help='Execute all universes')
@click.option('--thru', default=-1, help='Run until this universe number')
@click.option('--dir', 'folder', help='Multiverse directory',
              default='./multiverse', show_default=True)
def run(folder, run_all, num, thru):
    """ Execute the generated universe scripts.

    Run all universes: boba run --all

    Run a single universe, for example universe_1: boba run 1

    Run a range of universes for example 1 through 5: boba run 1 --thru 5
    """

    check_path(folder)

    if num < 0 and not run_all:
        print_help('Error: Missing argument "NUM".')

    cmd = ['sh', 'execute.sh']
    if not run_all:
        cmd.append(str(num))
        if thru > 0:
            cmd.append(str(thru))

    subprocess.run(cmd, cwd=folder)


@click.command()
@click.argument('pattern', nargs=1)
@click.option('--base', '-b', default='./multiverse/results',
              show_default=True, help='Folder containing the universe outputs')
@click.option('--out', default='./multiverse/merged.csv',
              show_default=True, help='Name of the merged file')
@click.option('--delimiter', default=',', show_default=True,
              help='CSV delimiter')
def merge(pattern, base, out, delimiter):
    """
    Merge CSV outputs from individual universes into one file.

    Required argument:
    the filename pattern of individual outputs where the universe id is
    replaced by {}, for example output_{}.csv
    """

    check_path(base)
    CSVMerger(pattern, base, out, delimiter).main()


@click.group()
@click.version_option()
def main():
    pass


main.add_command(compile)
main.add_command(run)
main.add_command(merge)

if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

import json
from dataclasses import dataclass
from .baseparser import ParseError
from .conditionparser import ConditionParser, TokenType


@dataclass
class Constraint:
    block: str = ''
    variable: str = ''
    option: str = ''
    index: int = -1
    skip: bool = False
    condition: str = ''


class ConstraintParser:
    def __init__(self, spec):
        self.spec = spec
        self.constraints = {}

        # useful for inferring ADGs
        self.links = []
        self.procedural = set()

    @staticmethod
    def _read_required(obj, field):
        if field not in obj:
            msg = 'Cannot find required field "{}"'.format(field)
            ConstraintParser._throw(msg, obj)
        return obj[field]

    @staticmethod
    def _read_optional(obj, field, df=None):
        return obj[field] if field in obj else df

    @staticmethod
    def _throw(msg, c):
        raise ParseError('In parsing constraints:\n\t' + json.dumps(c)
                         + '\n\t\t' + msg)

    @staticmethod
    def make_index_var(w):
        """ Embellish the variable name if the user is checking the option
         by its index in the options array."""
        return '_i_' + w

    @staticmethod
    def _verify_block(block, blocks, constraint):
        """ Check if a block is defined in the template script. """
        if block is not None and block not in blocks:
            msg = 'Block "{}" does not match any existing block ID'
            ConstraintParser._throw(msg.format(block), constraint)

    @staticmethod
    def _verify_placeholder_var(v, decs, constraint):
        """ Check if a placeholder variable is declared in decisions. """
        if v is not None and v not in decs:
            msg = 'Variable "{}" does not match any existing variable'
            ConstraintParser._throw(msg.format(v), constraint)

    @staticmethod
    def _verify_var_option(v, opt, decs, constraint):
        """ Check if an option to a placeholder variable exists. """
        opts = [str(o) for o in decs[v].value]
        if str(opt) not in opts:
            msg = 'Variable "{}" has no option "{}"'
            ConstraintParser._throw(msg.format(v, opt), constraint)

    @staticmethod
    def _verify_block_option(block, opt, bl_decs, constraint):
        """ Check if an option to a block exists."""
        if '{}:{}'.format(block, opt) not in bl_decs[block]:
            msg = 'Block "{}" has no option "{}"'
            ConstraintParser._throw(msg.format(block, opt), constraint)

    @staticmethod
    def _verify_option(param, block, opt, decs, bl_decs, c):
        """ Check if an option exists. """
        if opt is not None and param:
            ConstraintParser._verify_var_option(param, opt, decs, c)
        if opt is not None and block:
            ConstraintParser._verify_block_option(block, opt, bl_decs, c)

    @staticmethod
    def _verify_index(v, idx, decs, constraint):
        """ Check if the index of a placeholder variable is valid. """
        if idx is not None and (idx < 0 or idx >= len(decs[v].value)):
            msg = 'Index {} is out of range for variable "{}"'
            ConstraintParser._throw(msg.format(idx, v), constraint)

    @staticmethod
    def _verify_link(link, decs, bl_decs, constraint):
        """ Check if the linked decisions exist and have the same size """
        m = 0
        for l in link:
            if l not in decs and l not in bl_decs:
                msg = 'Decision "{}" not found'
                ConstraintParser._throw(msg.format(l), constraint)
            n = len(decs[l].value) if l in decs else len(bl_decs[l])
            if m != 0 and n != m:
                msg = 'Linked decisions must have the same number of options'
                ConstraintParser._throw(msg, constraint)
            m = n

    @staticmethod
    def _convert_link(link, decs, bl_decs):
        """ Convert linked decisions to constraints """
        res = []
        ls = []
        size = 0

        # first, turn variable/block decisions into a shared format
        for l in link:
            tp = 'variable' if l in decs else 'block'
            opt = decs[l].value if l in decs \
                else list(map(lambda x: x.split(':')[1], bl_decs[l]))
            ls.append({'type': tp, 'name': l, 'options': opt})
            size = len(opt)

        # then, construct pairwise dependencies
        for i in range(size):
            for l in ls:
                cond = ''
                for j in ls:
                    if j != l:
                        if j['type'] == 'variable':
                            cond += ' and ({}.index < 0 or {}.index == {})'\
                                .format(j['name'], j['name'], i)
                        else:
                            cond += ' and {} == {}'.format(j['name'], j['options'][i])
                cs = {l['type']: l['name'], 'condition': cond[5:], '_source': 'link'}
                if l['type'] == 'block':
                    cs['option'] = l['options'][i]
                else:
                    cs['index'] = i
                res.append(cs)
        return res

    @staticmethod
    def _verify_json_syntax(block, param, opt, constraint, idx):
        """ Check if the json has the correct combination of fields. """
        if block is not None and param is not None:
            msg = 'Cannot handle variable and block at the same line.'
            ConstraintParser._throw(msg, constraint)

        if opt is not None:
            if param is None and block is None:
                msg = 'No corresponding variable/block for option "{}"'
                ConstraintParser._throw(msg.format(opt), constraint)

        if param is not None:
            if opt is None and idx is None:
                msg = 'Must specify option/index for a variable.'
                ConstraintParser._throw(msg, constraint)

    @staticmethod
    def _verify_condition_vars(parsed, decs, blocks, constraint):
        """ Check if the variables and blocks in the condition exist. """
        for i, p in enumerate(parsed):
            if i % 2 == 1:
                # skipping the rhs
                continue

            # see if it matches anything from decs and blocks
            pv = p.value
            if pv not in decs and pv not in blocks:
                msg = '"{}" does not match any block or variable'.format(pv)
                ConstraintParser._throw(msg, constraint)

    @staticmethod
    def _maybe_get_index(param, opt, idx, decs):
        """ Placeholder's option is always checked by index. """
        if not param:
            return -1
        if idx is not None:
            return idx
        for i, o in enumerate(decs[param].value):
            if str(o) == str(opt):
                return i

    @staticmethod
    def _create_key(c):
        if c.block:
            key = '{}:{}'.format(c.block, c.option) if c.option else c.block
        else:
            v = ConstraintParser.make_index_var(c.variable)
            key = '{}:{}'.format(v, c.index)
        return key

    def _recon(self, code, parsed_decs, cond):
        """ Transform parsed code and decisions into valid python code """
        exe = []
        for i, d in enumerate(parsed_decs):
            if d.type == TokenType.index_var:
                exe.append(self.make_index_var(d.value))
            elif i % 2 == 1 and parsed_decs[i - 1].type == TokenType.var:
                # wrap any RHS in quotes, except for indices
                # because the history class represent any option as strings
                exe.append('"{}"'.format(d.value))
            else:
                exe.append(d.value)

        recon = code.format(*exe)

        # check if the code has syntax error
        try:
            eval(recon, {})
        except SyntaxError:
            msg = 'In parsing condition:\n\t' + cond + \
                  '\nSyntax Error: invalid syntax'
            raise ParseError(msg)
        except NameError:
            pass

        return recon

    def _infer_procedural_deps(self, c, block, variable, cond):
        """ Infer procedural edges from parsed condition """
        # skip constraints added by us, for example links
        if '_source' not in c:
            dec = block if block else variable
            # current node should depend on all decisions on the LHS
            for i in range(0, len(cond), 2):
                v = cond[i].value
                self.procedural.add('{}-{}'.format(v, dec))

    def read_constraints(self, code_parser, dec_parser):
        """ Read the constraints from the JSON spec. """
        cons = ConstraintParser._read_optional(self.spec, 'constraints', [])

        decs = dec_parser.decisions
        bls = code_parser.get_block_names()
        bl_decs = code_parser.get_decisions()

        # first separate links and conditions
        pure_cons = []
        for c in cons:
            # read link
            link = ConstraintParser._read_optional(c, 'link')
            if link:
                ConstraintParser._verify_link(link, decs, bl_decs, c)
                self.links.append(link)
                pure_cons += ConstraintParser._convert_link(link, decs, bl_decs)
            else:
                pure_cons.append(c)

        # add inline constraints
        pure_cons += code_parser.inline_constraints

        # then parse all conditions, including generated ones from link
        for c in pure_cons:
            # read block
            block = ConstraintParser._read_optional(c, 'block')
            ConstraintParser._verify_block(block, bls, c)

            # read variable
            param = ConstraintParser._read_optional(c, 'variable')
            ConstraintParser._verify_placeholder_var(param, decs, c)

            # read skip flag
            skip = bool(ConstraintParser._read_optional(c, 'skippable'))

            # read index
            idx = ConstraintParser._read_optional(c, 'index')
            ConstraintParser._verify_index(param, idx, decs, c)

            # read option
            opt = ConstraintParser._read_optional(c, 'option')
            ConstraintParser._verify_option(param, block, opt, decs, bl_decs, c)

            # verify everything and convert option to index for placeholder
            ConstraintParser._verify_json_syntax(block, param, opt, c, idx)
            idx = ConstraintParser._maybe_get_index(param, opt, idx, decs)

            # read condition
            cond = ConstraintParser._read_required(c, 'condition')
            code, parsed_decs = ConditionParser(cond).parse()
            if len(parsed_decs) % 2 == 1:
                ConstraintParser._throw('Binary operator expected', c)
            ConstraintParser._verify_condition_vars(parsed_decs, decs, bls, c)

            # get procedural dependency
            self._infer_procedural_deps(c, block, param, parsed_decs)

            # now transform it into valid python code
            recon = self._recon(code, parsed_decs, cond)

            # save
            constraint = Constraint(block, param, opt, idx, skip, recon)
            key = ConstraintParser._create_key(constraint)
            self.constraints[key] = constraint

        return self.constraints



# -*- coding: utf-8 -*-
import os

PY = 'python'
R = 'r'


class LangError(NameError):
    pass


class Lang:
    def __init__(self, lang, script):
        self.lang = self._infer_lang(lang, script)

    def _is_valid_lang(self, lang):
        return lang == PY or lang == R

    def _ext_to_lang(self, ext):
        ext = ext.lower()[1:]
        if ext == 'py':
            return PY
        if ext == 'r' or ext == 'rmd':
            return R
        return ''

    def _infer_lang(self, lang, script):
        lang = lang.strip().lower()

        if self._is_valid_lang(lang):
            return lang

        if lang == '':
            _, ext = os.path.splitext(script)
            lang = self._ext_to_lang(ext)
            if self._is_valid_lang(lang):
                return lang
            else:
                raise LangError('Error: cannot infer language from file ' +
                                'extension {}'.format(ext))
        else:
            raise LangError('Error: language "{}" is not supported'.format(lang))

    def get_ext(self):
        if self.lang == R:
            return '.R'
        if self.lang == PY:
            return '.py'

    def get_cmd(self):
        if self.lang == R:
            return 'Rscript'
        if self.lang == PY:
            return 'python'

    def is_r(self):
        return self.lang == R

    def is_python(self):
        return self.lang == PY



#!/usr/bin/env python3

# Ugly hack to allow import from the root folder
import sys
import os
sys.path.insert(0, os.path.abspath('..'))

import unittest
from boba.graphparser import GraphParser
from boba.graphanalyzer import GraphAnalyzer, InvalidGraphError


class TestGraphAnalyzer(unittest.TestCase):

    def cp_helper(self, spec, expected):
        nodes, edges = GraphParser(spec).parse()
        ga = GraphAnalyzer(nodes, edges)
        ga._construct_paths()
        expected = set([','.join(p) for p in expected])
        actual = set([','.join(p) for p in ga.paths])
        self.assertSetEqual(actual, expected)

    def test_construct_paths(self):
        # normal
        spec = ['a->b->c', 'b->d']
        expected = [['a', 'b', 'c'], ['a', 'b', 'd']]
        self.cp_helper(spec, expected)

        # single node
        spec = ['a']
        expected = [['a']]
        self.cp_helper(spec, expected)

        # multiple sources and targets
        spec = ['a->b->c', 'a2->b->c2']
        expected = [['a', 'b', 'c'], ['a2', 'b', 'c2'], ['a', 'b', 'c2'],
                    ['a2', 'b', 'c']]
        self.cp_helper(spec, expected)

        # disconnected
        spec = ['a->b->c', 'e->f']
        expected = [['a', 'b', 'c'], ['e', 'f']]
        self.cp_helper(spec, expected)

        # cyclic
        spec = ['a->b->c->a']
        nodes, edges = GraphParser(spec).parse()
        ga = GraphAnalyzer(nodes, edges)
        with self.assertRaises(InvalidGraphError):
            ga._construct_paths()

    def source_helper(self, spec, exp_source, exp_target):
        nodes, edges = GraphParser(spec).parse()
        ga = GraphAnalyzer(nodes, edges)
        self.assertSetEqual(ga._get_source(), exp_source)
        self.assertSetEqual(ga._get_target(), exp_target)

    def test_get_source_and_target(self):
        # normal
        spec = ['a->b->d', 'a->b->c']
        source = {'a'}
        target = {'d', 'c'}
        self.source_helper(spec, source, target)

        # disconnected
        spec = ['a->b->d', 'c->e']
        source = {'a', 'c'}
        target = {'d', 'e'}
        self.source_helper(spec, source, target)

        # a single node
        spec = ['a']
        source = {'a'}
        target = {'a'}
        self.source_helper(spec, source, target)

        # cyclic
        spec = ['a->b->c->d->a']
        source = set()
        target = set()
        self.source_helper(spec, source, target)

        # complex
        spec = ['a->b->d', 'a->c->b->d', 'c->a->d']
        source = set()
        target = {'d'}
        self.source_helper(spec, source, target)

    def path_helper(self, spec, s, t, expected):
        nodes, edges = GraphParser(spec).parse()
        ga = GraphAnalyzer(nodes, edges)
        ga._all_paths(s, t)
        expected = set([','.join(p) for p in expected])
        actual = set([','.join(p) for p in ga.paths])
        self.assertSetEqual(actual, expected)

    def test_get_path(self):
        """ test if the program correctly gets all paths from s to t"""

        spec = ['a->b->c', 'b->d', 'e']
        start = 'a'
        stop = 'e'
        expected = []
        self.path_helper(spec, start, stop, expected)

        spec = ['a->b->c', 'b->d', 'e']
        start = 'a'
        stop = 'c'
        expected = [['a', 'b', 'c']]
        self.path_helper(spec, start, stop, expected)

        spec = ['a->b->c', 'b->d', 'e']
        start = 'b'
        stop = 'd'
        expected = [['b', 'd']]
        self.path_helper(spec, start, stop, expected)

        # a single node
        spec = ['a']
        start = 'a'
        stop = 'a'
        expected = [['a']]
        self.path_helper(spec, start, stop, expected)

        # graph with a merged branch
        spec = ['a->b->c', 'b->d', 'c->e d->e']
        start = 'a'
        stop = 'e'
        expected = [['a', 'b', 'c', 'e'], ['a', 'b', 'd', 'e']]
        self.path_helper(spec, start, stop, expected)

        # graph with a cycle
        spec = ['a->b->c->b', 'b->d', 'e']
        start = 'a'
        stop = 'c'
        expected = [['a', 'b', 'c']]
        self.path_helper(spec, start, stop, expected)

        # a complicated graph
        spec = ['a->b->d', 'a->c->b->d', 'c->a->d']
        start = 'c'
        stop = 'd'
        expected = [['c', 'b', 'd'], ['c', 'a', 'd'], ['c', 'a', 'b', 'd']]
        self.path_helper(spec, start, stop, expected)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3

# Ugly hack to allow import from the root folder
import sys
import os
sys.path.insert(0, os.path.abspath('..'))

import unittest
from unittest.mock import patch
import io
from boba.parser import Parser

FIRST_SCRIPT = 'multiverse/code/universe_1.py'


def _print_code(ps):
    for b in ps.blocks:
        bl = ps.blocks[b]
        print('-' * 10, '  ({}) {}  '.format(bl.id, bl.name), '-' * 10)
        for c in bl.chunks:
            print(c[1], end='')


def abs_path(rel_path):
    return os.path.join(os.path.dirname(__file__), rel_path)


class TestParser(unittest.TestCase):

    # --- code gen ---
    # a simple synthetic example
    def test_code_gen(self):
        base = abs_path('../example/simple/')
        ps = Parser(base+'template.py', base)
        ps.main(verbose=False)
        self.assertEqual(ps.wrangler.counter, 6)

    # a complex example
    def test_codegen_reading(self):
        base = abs_path('../example/reading/python/')
        Parser(base+'template.py', base).main(verbose=False)

    # another complex example
    def test_codegen_fertility(self):
        base = abs_path('../example/fertility/')
        ps = Parser(base+'template.py', base)
        ps.main(verbose=False)
        self.assertEqual(ps.wrangler.counter, 120)

    # an example written in R
    def test_r(self):
        base = abs_path('../example/fertility_r/')
        ps = Parser(base+'template.R', base)
        ps.main(verbose=False)
        self.assertEqual(ps.wrangler.counter, 120)

    # the spec has one decision and no graphs; should work
    def test_codegen_decision_only(self):
        base = abs_path('./specs/')
        sc = 'script-no-graph.py'
        ps = Parser(base+sc, base)
        ps.main(verbose=False)

        # compare file content
        with open(base+FIRST_SCRIPT, 'r') as f:
            actual = f.read()
        with open(base+sc, 'r') as f:
            lines = f.readlines()[7:]
            expected = ''.join(lines).replace('{{a}}', '1')
        self.assertEqual(actual, expected)

    # the spec has a graph and no decisions; should work
    def test_codegen_graph_only(self):
        base = abs_path('./specs/')
        ps = Parser(base + 'script1-good.py', base)
        ps.main(verbose=False)

        # compare file content
        with open(os.path.join(base, FIRST_SCRIPT), 'r') as f:
            actual = f.read()
        with open(base+'script1.py', 'r') as f:
            lines = f.read().split('\n')
            expected = ''
            for l in lines:
                if not l.strip().startswith('# --- '):
                    expected += l + '\n'
        self.assertEqual(actual.strip(), expected.strip())

    # the template contains decisions, but spec does not have corresponding def
    @patch('sys.stdout', new_callable=io.StringIO)
    def test_codegen_missing_decision(self, stdout):
        base = abs_path('./specs/')
        sc = 'script-no-graph-empty.py'
        with self.assertRaises(SystemExit):
            Parser(base+sc, base)
        self.assertRegex(stdout.getvalue(), 'Cannot find matching variable')

    # --- parse blocks ---
    def test_parse_blocks(self):
        base = abs_path('../example/simple/')
        ps = Parser(base+'template.py')
        self.assertSetEqual(set(ps.code_parser.blocks.keys()), {'_start', 'A:std', 'A:iqr', 'B'})
        self.assertListEqual(['_start', 'A', 'B'], ps.code_parser.order)

    def test_script_2(self):
        ps = Parser(abs_path('./specs/')+'script2.py')
        self.assertListEqual([*ps.code_parser.blocks], ['_start', 'A', 'B', 'C'])

    @patch('sys.stdout', new_callable=io.StringIO)
    def test_script_3(self, stdout):
        with self.assertRaises(SystemExit):
            Parser(abs_path('./specs/') + 'script2-syntax.py')
        self.assertRegex(stdout.getvalue(), r'Cannot find "\("')

    @patch('sys.stdout', new_callable=io.StringIO)
    def test_script_4(self, stdout):
        with self.assertRaises(SystemExit):
            Parser(abs_path('./specs/')+'script2-dup.py')
        self.assertRegex(stdout.getvalue(), '(?i)duplicated')

    @patch('sys.stdout', new_callable=io.StringIO)
    def test_script_5(self, stdout):
        with self.assertRaises(SystemExit):
            Parser(abs_path('./specs/')+'script2-dup-var.py')
        self.assertRegex(stdout.getvalue(), '(?i)name')

    def test_script_6(self):
        base = abs_path('./specs/')
        ps = Parser(base+'script2-block-param.py', base)
        ps.main(verbose=False)
        self.assertEqual(ps.wrangler.counter, 9)

    # --- constraints ---
    def test_constraint_1(self):
        """ Block options depend on block parameter """
        base = abs_path('./specs/')
        ps = Parser(base+'script3-1.py', base)
        ps.main(verbose=False)
        self.assertEqual(ps.wrangler.counter, 6)

    def test_constraint_2(self):
        """ Block parameter depends on block parameter """
        base = abs_path('./specs/')
        ps = Parser(base+'script3-2.py', base)
        ps.main(verbose=False)
        self.assertEqual(ps.wrangler.counter, 3)

    def test_constraint_3(self):
        """ Normal block depends on block parameter """
        base = abs_path('./specs/')
        ps = Parser(base+'script3-3.py', base)
        ps.main(verbose=False)
        self.assertEqual(ps.wrangler.counter, 8)

    def test_constraint_4(self):
        """ Variable depends on variable """
        base = abs_path('./specs/')
        ps = Parser(base + 'script3-4.py', base)
        ps.main(verbose=False)
        self.assertEqual(ps.wrangler.counter, 4)

        # then, test index
        ps = Parser(base + 'script3-4.py', base)
        ps.spec['constraints'] = [
            {"variable": "b", "index": 1, "condition": "a.index == 0"},
            {"variable": "b", "index": 0, "condition": "a == else"}]
        ps._parse_constraints()
        ps.main(verbose=False)
        self.assertEqual(ps.wrangler.counter, 4)

    def test_constraint_5(self):
        """ Skip a block """
        # first, skip a normal block
        base = abs_path('./specs/')
        ps = Parser(base+'script3-5.py', base)
        ps.main(verbose=False)
        self.assertEqual(ps.wrangler.counter, 8)

        # then, skip a decision block
        ps = Parser(base+'script3-5.py', base)
        ps.spec['constraints'] = [{"block": "B", "skippable": True, "condition": "a == if"}]
        ps._parse_constraints()
        ps.main(verbose=False)
        self.assertEqual(ps.wrangler.counter, 6)

    def test_constraint_7(self):
        """ Linked decisions """
        base = abs_path('./specs/')
        ps = Parser(base+'script3-7.py', base)
        ps.main(verbose=False)
        self.assertEqual(ps.wrangler.counter, 10)

    def test_constraint_inline(self):
        """ Inline constraints """
        base = abs_path('./specs/')
        ps = Parser(base+'script-inline-constraints.py', base)
        ps.main(verbose=False)
        self.assertEqual(ps.wrangler.counter, 2)

    # --- adg ---
    @staticmethod
    def _edge_to_set(edges):
        res = {}
        for k in edges:
            res[k] = set(edges[k])
        return res

    def test_adg(self):
        base = abs_path('./specs/')
        ps = Parser(base+'script4-1.py', base)
        ps.adg.create(ps.code_parser.blocks)
        self.assertSetEqual(ps.adg.nodes, {'b', 'C', 'B', 'A'})
        expected = {'B': {'b', 'C'}, 'A': {'B'}}
        self.assertDictEqual(self._edge_to_set(ps.adg.edges), expected)

    def test_adg_code_graph(self):
        base = abs_path('./specs/')
        ps = Parser(base+'script4-2.py', base)
        ps.adg.create(ps.code_parser.blocks)
        self.assertSetEqual(ps.adg.nodes, {'C', 'B', 'a', 'A', 'b', 'D'})
        expected = {'b': {'D', 'C'}, 'B': {'D', 'b', 'C'}, 'a': {'B'}, 'A': {'a'}}
        self.assertDictEqual(self._edge_to_set(ps.adg.edges), expected)
        expected ={'b': {'D', 'C'}, 'B': {'D', 'b', 'C'}}
        self.assertDictEqual(self._edge_to_set(ps.adg.proc_edges), expected)

    def test_adg_link(self):
        base = abs_path('./specs/')
        ps = Parser(base+'script4-3.py', base)
        ps.adg.create(ps.code_parser.blocks)
        self.assertSetEqual(ps.adg.nodes, {'a', 'B'})
        self.assertDictEqual(ps.adg.edges, {'a': ['B']})

    # --- parse graph ---
    def test_spec_good(self):
        base = abs_path('../example/simple/')
        ps = Parser(base+'template.py')
        expected = [['_start', 'A:std', 'B'], ['_start', 'A:iqr', 'B']]
        expected = set([','.join(p) for p in expected])
        self.assertSetEqual(set([','.join(p) for p in ps.paths]), expected)

    def test_spec_empty(self):
        base = abs_path('./specs/')
        ps = Parser(base+'script1.py')
        self.assertListEqual(ps.paths, [['A', 'B', 'C']])

    @patch('sys.stdout', new_callable=io.StringIO)
    def test_spec_bad_graph(self, stdout):
        base = abs_path('./specs/')
        with self.assertRaises(SystemExit):
            Parser(base+'script1-bad-graph.py')
        self.assertRegex(stdout.getvalue(), 'Cannot find a target node')

    @patch('sys.stdout', new_callable=io.StringIO)
    def test_spec_cyclic_graph(self, stdout):
        base = abs_path('./specs/')
        with self.assertRaises(SystemExit):
            Parser(base + 'script1-cyclic-graph.py')
        self.assertRegex(stdout.getvalue(), 'Cannot find any starting node')


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

from dataclasses import dataclass, field
from typing import List
import json

from .blocksyntaxparser import BlockSyntaxParser, ParseError


@dataclass
class Block:
    """
    A class for code blocks.

    id: unique identifier. For decision block, id is parameter:option.
    parameter: parameter name, if the block is a decision.
    option: option name, if the block is a decision.
    chunks: code broken up at the boundaries of placeholder variables.
    """

    id: str = ''
    parameter: str = ''
    option: str = ''
    chunks: List = field(default_factory=lambda: [])


@dataclass
class Chunk:
    """A class for code chunks.
    A code chunk contains at most one placeholder variable.

    variable: the corresponding placeholder variable, if any.
    code: the code template proceeding the variable or the block boundary.
    """
    variable: str = ''
    code: str = ''


class CodeParser:
    def __init__(self):
        self.blocks = {}
        self.order = []

        self.raw_spec = ''
        self.spec = {}

        self.inline_constraints = []
        self.inline_vars = []
        self.used_vars = set()

    @staticmethod
    def _get_block_name(block):
        """Get the ID of the block, ignoring options."""
        return block.id if block.parameter == '' else block.parameter

    def _add_block(self, block):
        """Add a block to our data structure."""
        # handle config block
        if block.id == 'BOBA_CONFIG':
            self.raw_spec += block.chunks[0].code
            return
        if block.id == 'END':
            block.id = ''
            if len(self.order):
                self.blocks[self.order[-1]].chunks += block.chunks
                return

        # ignore empty block
        if block.id == '' and block.chunks[0].code == '':
            return

        # handle unnamed block
        if block.id == '':
            block.id = '_start' if len(self.blocks) == 0 else '_end'

        # check if id exists
        if block.id in self.blocks:
            err = 'Duplicated code block ID "{}"'.format(block.id)
            raise ParseError(err)

        # add to data structure
        self.blocks[block.id] = block
        bn = CodeParser._get_block_name(block)
        if bn not in self.order:
            self.order.append(bn)

    def get_block_names(self):
        """
        Get the ID of all blocks, ignoring options
        :return: a set of unique names
        """
        blocks = set()
        for b in self.blocks:
            bl = self.blocks[b]
            blocks.add(CodeParser._get_block_name(bl))
        return blocks

    def get_decisions(self):
        """
        Get a dict of all block-level decisions, where the key is the parameter
        and the value is a list of block ids (namely, parameter:option).
        :return:
        """
        decs = {}
        for b in self.blocks:
            bl = self.blocks[b]
            if bl.parameter:
                p = bl.parameter
                if p in decs:
                    decs[p].append(bl.id)
                else:
                    decs[p] = [bl.id]
        return decs

    def parse(self, dec_parser, f):
        """ Make a pass over the template, parsing block declarations and
        placeholder variables inside the code."""
        code = ''
        bl = Block()

        for line in f:
            if BlockSyntaxParser.can_parse(line):
                # end of the previous block
                bl.chunks.append(Chunk('', code))
                code = ''
                self._add_block(bl)

                # parse the metadata and create a new block
                bp_id, par, opt, cond = BlockSyntaxParser(line).parse()
                bl = Block(bp_id, par, opt, [])

                # store inline constraints, if any
                if cond:
                    self.inline_constraints.append(cond)
            else:
                # match decision variables
                try:
                    vs, codes = dec_parser.parse_code(line)
                    if len(vs):
                        # store inline variables
                        self.used_vars.update(vs)

                        # chop into more chunks
                        # combine first chunk with previous code
                        bl.chunks.append(Chunk(vs[0], code + codes[0]))
                        for i in range(1, len(vs)):
                            bl.chunks.append(Chunk(vs[i], codes[i]))

                        # remaining code after the last matched variable
                        code = codes[-1]
                    else:
                        code += line
                except ParseError as e:
                    msg = 'At line "{}"\n\t{}'.format(line, e.args[0])
                    raise ParseError(msg)

        # add the last block
        bl.chunks.append(Chunk('', code))
        self._add_block(bl)

        # parse the spec
        try:
            self.spec = json.loads(self.raw_spec) if self.raw_spec else {}
        except ValueError as e:
            msg = self.raw_spec + '\n' + e.args[0]
            msg += '\nBoba config is not valid JSON'
            raise ParseError(msg)



# -*- coding: utf-8 -*-

from .baseparser import BaseParser, ParseError

kw = '# ---'


class BlockSyntaxParser(BaseParser):
    """
    Parse the metadata of a code block, which must have the structure:
        # --- (ID) option @if condition
    option is optional, but including it will mark the block as a parameter.
    @if is optional and it creates a procedural dependency constraint on this
        block and this option (if any).
    """

    def __init__(self, line):
        super(BlockSyntaxParser, self).__init__(line)

        self.state = 0
        self.parsed_id = ''
        self.parsed_parameter = ''
        self.parsed_option = ''
        self.parsed_condition = ''

    @staticmethod
    def can_parse(line):
        return line.lstrip().startswith(kw)

    @staticmethod
    def _is_operator_start(ch):
        return ch == '@'

    @staticmethod
    def _is_condition(word):
        return word == 'if'

    def parse(self):
        while not self._is_end():
            self._read_next()
        return self.parsed_id, self.parsed_parameter, self.parsed_option,\
            self.parsed_condition

    def _read_next(self):
        self._read_while(BlockSyntaxParser._is_whitespace)
        if self._is_end():
            return

        if self.state == 0:
            self._read_kw()
        elif self.state == 1:
            self._read_id()
        elif self.state == 2:
            self._maybe_read_option()
        elif self.state == 3:
            self._read_condition()
        else:
            # we've read anything we can handle but haven't reached the end
            s = self._remaining().strip()
            self._throw('Cannot handle "{}"'.format(s))

    def _end(self):
        self.i = len(self.line)  # stop parsing

    def _throw(self, msg):
        err = 'At character {} of "{}":\n\t{}'.format(self.i+1, self.line, msg)
        raise ParseError(err)

    def _remaining(self):
        return self.line[self.i:]

    def _read_kw(self):
        if self._remaining().startswith(kw):
            self.i += len(kw)
            self.state += 1
        else:
            self._throw('expected {}'.format(kw))

    def _read_id(self):
        """ Read the thing inside the parenthesis. """
        # open paren
        if self._peek_char() != '(':
            self._throw('Cannot find "("')
        self._next_char()
        self._read_while(self._is_whitespace)

        # read the actual identifier
        ch = self._peek_char()
        if not self._is_id_start(ch):
            self._throw('Invalid identifier start character {}'.format(ch))

        self.parsed_id = self._read_while(self._is_id)
        self._read_while(self._is_whitespace)

        # close paren
        if self._peek_char() != ')':
            self._throw('Cannot find ")"')
        self._next_char()
        self.state += 1

    def _maybe_read_option(self):
        """ Read the option, if there is any. """
        self._read_while(self._is_whitespace)

        # check if the next word is maybe an option
        if not self._is_id_start(self._peek_char()):
            self.state += 1
            return

        # option follows the same naming convention as ID
        opt = self._read_while(self._is_id)
        if opt != '':
            self.parsed_parameter = self.parsed_id
            self.parsed_option = opt
            self.parsed_id += ':' + self.parsed_option

        self._read_while(self._is_whitespace)
        self.state += 1

    def _read_condition(self):
        """ Read condition. """
        self._read_while(self._is_whitespace)

        # check if the next char is indeed an operator
        if not BlockSyntaxParser._is_operator_start(self._peek_char()):
            self.state += 1
            return

        # read @if
        self._next_char()
        w = self._read_while(self._is_id)
        if not BlockSyntaxParser._is_condition(w):
            self._throw('Cannot handle @{}'.format(w))

        # read whatever remains as the condition
        s = self._remaining().strip()
        self._end()
        self.state += 1

        # construct the condition
        bl = self.parsed_parameter if self.parsed_option else self.parsed_id
        self.parsed_condition = {'block': bl, 'condition': s}
        if self.parsed_option:
            self.parsed_condition['option'] = self.parsed_option



#!/usr/bin/env python3

# Ugly hack to allow import from the root folder
import sys
import os
sys.path.insert(0, os.path.abspath('..'))

import unittest
import json
from boba.decisionparser import DecisionParser


def abs_path(rel_path):
    return os.path.join(os.path.dirname(__file__), rel_path)


class TestDecisionParser(unittest.TestCase):

    def test_id_syntax(self):
        # valid identifiers
        self.assertTrue(DecisionParser._is_id_token('my_var'))
        self.assertTrue(DecisionParser._is_id_token('A1'))
        self.assertTrue(DecisionParser._is_id_token('a1b'))

        # invalid identifiers
        self.assertFalse(DecisionParser._is_id_token('_start'))
        self.assertFalse(DecisionParser._is_id_token('1b'))
        self.assertFalse(DecisionParser._is_id_token(' A'))

    def test_read_json(self):
        with open(abs_path('./specs/spec-good.json'), 'rb') as f:
            spec = json.load(f)
        dp = DecisionParser()
        ds = dp.read_decisions(spec)
        self.assertListEqual(list(ds.keys()), ['a', 'b'])
        self.assertEqual(ds['a'].desc, 'outlier')
        self.assertEqual(ds['b'].desc, 'Decision b')

    def test_parse_variable_def(self):
        # numbers
        dp = DecisionParser()
        line = 'a = b + {{c = 1, 2}}'
        vs, codes = dp.parse_code(line)
        self.assertListEqual(vs, ['c'])
        self.assertListEqual(codes, ['a = b + ', ''])
        self.assertListEqual(dp.decisions['c'].value, [1, 2])

        # strings
        line = 'family = {{fml="lognormal","normal"}}()'
        vs, codes = dp.parse_code(line)
        self.assertListEqual(vs, ['fml'])
        self.assertListEqual(codes, ['family = ', '()'])
        self.assertListEqual(dp.decisions['fml'].value, ["lognormal", "normal"])

    def test_parse_code(self):
        with open(abs_path('./specs/spec-good.json'), 'rb') as f:
            spec = json.load(f)
        dp = DecisionParser()
        dp.read_decisions(spec)

        line = ''
        vs, codes = dp.parse_code(line)
        self.assertListEqual(vs, [])
        self.assertListEqual(codes, [line])

        # valid pattern, no variable
        line = '{{}}'
        vs, codes = dp.parse_code(line)
        self.assertListEqual(vs, [])
        self.assertListEqual(codes, [line])

        # valid pattern {{a}}
        line = "\t this is {{a}} v{{a}}riable"
        vs, codes = dp.parse_code(line)
        self.assertListEqual(vs, ['a', 'a'])
        self.assertListEqual(codes, ['\t this is ', ' v', 'riable'])

        # invalid id start {{_a}}
        line = '{{_a}}'
        vs, codes = dp.parse_code(line)
        self.assertListEqual(vs, [])

        # valid pattern, back to back
        line = '{{a}}{{b}}'
        vs, codes = dp.parse_code(line)
        self.assertListEqual(vs, ['a', 'b'])
        self.assertListEqual(codes, ['', '', ''])

        # back to back, too few separators
        line = '{{a}}{a}}'
        vs, codes = dp.parse_code(line)
        self.assertListEqual(vs, ['a'])
        self.assertListEqual(codes, ['', '{a}}'])

        # back to back, extra separators
        line = '{{a}}{{{b}}'
        vs, codes = dp.parse_code(line)
        self.assertListEqual(vs, ['a', 'b'])
        self.assertListEqual(codes, ['', '{', ''])

        # back to back, extra separators
        line = '{{a}}}{{b}}'
        vs, codes = dp.parse_code(line)
        self.assertListEqual(vs, ['a', 'b'])
        self.assertListEqual(codes, ['', '}', ''])

        # broken + valid
        line = '{{a}{{a}}'
        vs, codes = dp.parse_code(line)
        self.assertListEqual(vs, ['a'])
        self.assertListEqual(codes, ['{{a}', ''])

        # broken + valid
        line = '{{{{a}}'
        vs, codes = dp.parse_code(line)
        self.assertListEqual(vs, ['a'])
        self.assertListEqual(codes, ['{{', ''])

        # no pattern
        line = "'In parsing file \"{}\":\n'.format(self.fn_script)"
        vs, codes = dp.parse_code(line)
        self.assertListEqual(vs, [])
        self.assertListEqual(codes, [line])

        # missing closing syntax
        line = '{{a}'
        vs, codes = dp.parse_code(line)
        self.assertListEqual(vs, [])
        self.assertListEqual(codes, [line])

        # missing closing syntax
        line = '{{a'
        vs, codes = dp.parse_code(line)
        self.assertListEqual(vs, [])
        self.assertListEqual(codes, [line])


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

import os
import shutil
import csv
import json
from dataclasses import dataclass
from .baseparser import ParseError


@dataclass
class Output:
    name: str
    value: str


exec_template = """\
#!/bin/sh

cleanup ()
{{
echo "kill -s TERM $!"
kill -s TERM $!
exit 0
}}

merge_log ()
{{
boba merge exit_status_{{}}.csv -b $DLOG --out $DLOG/exit_status.csv
rm $DLOG/exit_status_*.csv
}}

{}

DIR="$( cd "$( dirname "${{BASH_SOURCE[0]}}" )" >/dev/null 2>&1 && pwd )"
DLOG=$DIR/boba_logs
suffix={}
num={}
i=1

# create a folder for logs
rm -rf $DLOG
mkdir $DLOG

# if specified, change the range of universes to execute
if [ "$1" != "" ]; then i=$1; num=$1; fi
if [ "$2" != "" ]; then num=$2; fi

cd $DIR/code

trap cleanup SIGINT SIGTERM

while [ $i -le $num ]
do
  # execute the universe
  f="universe_$i$suffix"
  echo "{} $f"
  (set -o pipefail; {} $f 2>&1 | tee $DLOG/log_$i.txt)

  # collect exit status
  printf "%s\\n%d" exit_status $? > $DLOG/exit_status_$i.csv

  # increment
  i=$(( i+1 ))
  wait $!
done

merge_log
{}
"""

DIR_SCRIPT = 'code/'


class Wrangler:
    """Handles outputs."""
    def __init__(self, spec, lang, out):
        self.spec = spec
        self.lang = lang
        self.out = out
        self.fn = os.path.abspath(os.path.join(out, 'summary.csv'))

        self.outputs = {}
        self.col = 0  # output column number, will be set by parser
        self.counter = 0

        self.pre_exe = ''
        self.post_exe = ''

        self._read_spec()

    @staticmethod
    def _read_json_safe(obj, field):
        if field not in obj:
            raise ParseError('Cannot find "{}" in json'.format(field))
        return obj[field]

    @staticmethod
    def _read_optional(obj, field, df):
        return obj[field] if field in obj else df

    def _read_spec(self):
        """Read misc fields from the JSON spec."""
        sp = self._read_optional(self.spec, 'outputs', [])
        for d in sp:
            name = str(self._read_json_safe(d, 'name'))
            value = str(self._read_json_safe(d, 'value'))
            o = Output(name, value)
            self.outputs[name] = o

        self.pre_exe = self._read_optional(self.spec, 'before_execute', '')
        self.post_exe = self._read_optional(self.spec, 'after_execute', '')

    def _codegen_r(self):
        """Generate output code for R scripts."""
        if len(self.outputs) == 0:
            return ''

        # read csv
        code = '\n\n# wrangles output\n' \
            'df <- read.csv2("{}", sep = ",", stringsAsFactors = FALSE, ' \
               'check.names=FALSE)'\
            .format(self.fn)

        # record outputs
        ns = self.get_outputs()
        col = self.col + 1
        row = self.counter
        for n in ns:
            code += '\ndf[{}, {}] = {}'.format(row, col, self.outputs[n].value)
            col += 1

        # write csv
        code += '\nwrite.csv(df, file="{}", row.names=FALSE)'.format(self.fn)
        code += '\n'

        return code

    def _codegen_python(self):
        if len(self.outputs) == 0:
            return ''

        # TODO

    def _gen_code(self):
        """Generate output code to be appended to the end of the script."""
        if self.lang.is_r():
            return self._codegen_r()
        if self.lang.is_python():
            return self._codegen_python()

    def write_sh(self):
        """Write a shell script for executing all universes."""
        cmd = self.lang.get_cmd()
        sh = exec_template.format(self.pre_exe,
                                  self.lang.get_ext(), self.counter, cmd, cmd,
                                  self.post_exe)
        fn_exec = os.path.join(self.out, 'execute.sh')
        with open(fn_exec, 'w') as f:
            f.write(sh)
        st = os.stat(fn_exec)
        os.chmod(fn_exec, st.st_mode | 0o0111)

    def write_universe(self, code):
        """Write the generated code to a universe file."""

        self.counter += 1
        fn = 'universe_{}{}'.format(self.counter, self.lang.get_ext())

        # replace the reserved keyword _n
        code = code.replace('{{_n}}', str(self.counter))

        # append output code
        code += self._gen_code()

        # write file
        with open(os.path.join(self.out, DIR_SCRIPT, fn), 'w') as f:
            f.write(code)
            f.flush()

        return fn

    def write_summary(self, rows):
        """Write the summary CSV file"""
        with open(self.fn, 'w', newline='') as f:
            wrt = csv.writer(f)
            for row in rows:
                wrt.writerow(row)

    def write_overview_json(self, res):
        """ Write the overview.json file"""
        # append visualizer block
        default_config = {
            "files": [{"id": "est", "path": "estimates.csv"}],
            "schema": {"point_estimate": {"file": "est", "field": "estimate"}}
        }
        vis = Wrangler._read_optional(self.spec, 'visualizer', None)

        # if it is a string, read config file
        if isinstance(vis, str):
            try:
                with open(vis) as f:
                    vis = json.load(f)
            except (IOError, json.JSONDecodeError) as e:
                print(e)
                print('Cannot read the visualizer config, using the default')
                vis = default_config
        # if user does not specify the config, use the default
        vis = default_config if vis is None else vis
        res['visualizer'] = vis

        with open(os.path.join(self.out, 'overview.json'), 'w') as f:
            obj = json.dumps(res, indent=2, sort_keys=True)
            f.write(obj)

    def create_dir(self):
        """Create output directories."""
        if os.path.exists(self.out):
            shutil.rmtree(self.out)
        os.makedirs(self.out)
        os.makedirs(os.path.join(self.out, DIR_SCRIPT))

    def get_outputs(self):
        """Get a sorted list of output names."""
        return sorted(list(self.outputs.keys()))



# -*- coding: utf-8 -*-


class Colors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


def print_fail(msg):
    print(Colors.FAIL + msg + Colors.ENDC)

def print_warn(msg):
    print(Colors.WARNING + msg + Colors.ENDC)



# -*- coding: utf-8 -*-

import re
import sys
from dataclasses import dataclass


@dataclass
class Token:
    type: str
    value: str


class ParseError(SyntaxError):
    pass


class BaseParser:

    def __init__(self, line):
        self.line = line
        self.i = 0
        self.row = 0
        self.col = 0
        self.current = None

    @staticmethod
    def _is_whitespace(char):
        return any(c == char for c in ' \t\n')

    @staticmethod
    def _is_id_start(ch):
        return bool(re.match('[a-zA-Z]', ch))

    @staticmethod
    def _is_id(ch):
        return bool(re.match('[_a-zA-Z0-9]', ch))

    @staticmethod
    def _is_digit(ch):
        return bool(re.match('[0-9]', ch))

    def _next_char(self):
        ch = self.line[self.i]
        self.i += 1
        if ch == '\n':
            self.row += 1
            self.col = 0
        else:
            self.col += 1
        return ch

    def _peek_char(self):
        return self.line[self.i]

    def _is_end(self):
        return self.i >= len(self.line)

    def _read_while(self, fun, max_len=sys.maxsize):
        s = ''
        while not self._is_end() and fun(self._peek_char()) and len(s) < max_len:
            s += self._next_char()
        return s

    def _peek(self):
        if not self.current:
            self.current = self._read_next()
        return self.current

    def _next(self):
        tmp = self.current
        self.current = None
        return tmp or self._read_next()

    def _read_next(self):
        pass



# -*- coding: utf-8 -*-


class InvalidGraphError(NameError):
    pass


class GraphAnalyzer:
    def __init__(self, nodes, edges):
        self.nodes = nodes
        self.edges = GraphAnalyzer._convert_edges(edges)
        self.paths = []

    @staticmethod
    def _convert_edges(edges):
        d = {}
        for e in edges:
            if e.start in d:
                d[e.start].append(e.end)
            else:
                d[e.start] = [e.end]
        return d

    def _throw(self, msg):
        msg = 'In analyzing graph structure:\n\t' + msg
        raise InvalidGraphError(msg)

    def _all_ending_nodes(self):
        """ nodes that have at least one incoming edge(s) """
        flat = [item for lst in self.edges.values() for item in lst]
        return set(flat)

    def _get_source(self):
        """ nodes that have no incoming edges """
        return self.nodes.difference(self._all_ending_nodes())

    def _get_target(self):
        """ nodes that have no outgoing edges """
        return self.nodes.difference(set(self.edges.keys()))

    def _all_paths_recur(self, a, b, visited, path):
        """ a recursive func to get all paths from a to b"""
        # mark the current node as visited and add to path
        visited.add(a)
        path.append(a)

        # if current node is the same as target, the path is done
        if a == b:
            self.paths.append([nd for nd in path])
        else:
            if a in self.edges:
                for n in self.edges[a]:
                    if n not in visited:
                        self._all_paths_recur(n, b, visited, path)

        # remove current node from path and mark it as unvisited
        path.pop()
        visited.discard(a)

    def _all_paths(self, s, t):
        """ get all paths from s to t """
        visited = set()
        path = []
        self._all_paths_recur(s, t, visited, path)

    def _construct_paths(self):
        ss = self._get_source()
        ts = self._get_target()

        if len(ss) == 0:
            self._throw('Cannot find any starting node')
        if len(ts) == 0:
            self._throw('Cannot find any ending node')

        for s in ss:
            for t in ts:
                self._all_paths(s, t)

    def analyze(self):
        if len(self.nodes) == 0:
            return []

        self._construct_paths()
        return self.paths



# -*- coding: utf-8 -*-
from .baseparser import BaseParser, ParseError
from dataclasses import dataclass
from enum import Enum


class TokenType(Enum):
    var = 1
    index_var = 2
    number = 3


@dataclass
class ParsedToken:
    value: str
    type: TokenType


class ConditionParser(BaseParser):
    """ A class for parsing the condition string """
    def __init__(self, line):
        super(ConditionParser, self).__init__(line)
        self.parsed_code = ''
        self.parsed_decs = []

    def parse(self):
        while not self._is_end():
            self._read_next()
        return self.parsed_code, self.parsed_decs

    @staticmethod
    def _is_keyword(w):
        return w == 'and' or w == 'or'

    @staticmethod
    def _is_operator(ch):
        return ch in ['=', '(', ')', '!', '>', '<']

    def _throw(self, msg):
        msg = 'At character {} of "{}":\n\t{}'.format(self.i + 1, self.line, msg)
        raise ParseError(msg)

    def _maybe_read_index(self):
        # we only want to parse the LHS of ==
        if len(self.parsed_decs) % 2 == 1:
            return False

        if not self._is_end() and self._peek_char() == '.':
            # try to parse .index
            self._next_char()
            v = self._read_while(self._is_id)
            if v == 'index':
                return True
            else:
                msg = 'Expected ".index", got ".{}"'.format(v)
                self._throw(msg)

        return False

    def _read_next(self):
        self.parsed_code += self._read_while(BaseParser._is_whitespace)
        if self._is_end():
            return

        ch = self._peek_char()
        if self._is_id_start(ch):
            w = self._read_while(self._is_id)
            if ConditionParser._is_keyword(w):
                self.parsed_code += w
                return

            tk = ParsedToken(w, TokenType.var)
            if self._maybe_read_index():
                tk.type = TokenType.index_var

            self.parsed_decs.append(tk)
            self.parsed_code += '{}'
        elif self._is_digit(ch):
            w = self._read_while(self._is_digit)
            if not self._is_end() and self._peek_char() == '.':  # read decimal
                w += self._next_char() + self._read_while(self._is_digit)

            self.parsed_decs.append(ParsedToken(w, TokenType.number))
            self.parsed_code += '{}'
        elif self._is_operator(ch):
            w = self._read_while(ConditionParser._is_operator)
            self.parsed_code += w
        else:
            msg = 'Cannot handle character "{}".'.format(ch)
            self._throw(msg)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

with open("README.rst", "r") as fh:
    readme = fh.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = ['Click>=6.0', 'dataclasses>=0.6', 'pandas>=0.24.2']

setup_requirements = []

test_requirements = []

setup(
    author="Yang Liu",
    author_email='yliu0@uw.edu',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    description="Author and execute multiverse analysis",
    entry_points={
        'console_scripts': [
            'boba=boba.cli:main',
        ],
    },
    install_requires=requirements,
    license="BSD license",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='multiverse analysis',
    name='boba',
    packages=find_packages(include=['boba', 'boba.*']),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/uwdata/multiverse-spec',
    version='0.1.5',
    zip_safe=False,
)




class ADG:
    """ For creating ADG. """
    def __init__(self):
        self.nodes = set()
        self.edges = {}
        self.proc_edges = {}

        self._graph_nodes = set()
        self._graph_edges = {}
        self._links = []    # linked decisions
        self._constraint_proc = set()  # procedural deps from constraints
        self._decs = set()  # all decisions

    @staticmethod
    def _convert_edges(edges):
        d = {}
        for e in edges:
            ADG._add_edge(d, e.start, e.end)
        return d

    @staticmethod
    def _add_edge(res, start, end):
        if start in res and end not in res[start]:
            res[start].append(end)
        else:
            res[start] = [end]

    @staticmethod
    def _all_ending_nodes(edges):
        """ nodes that have at least one incoming edge(s) """
        flat = [item for lst in edges.values() for item in lst]
        return set(flat)

    @staticmethod
    def _get_source(nodes, edges):
        """ nodes that have no incoming edges """
        return nodes.difference(ADG._all_ending_nodes(edges))

    @staticmethod
    def _get_target(nodes, edges):
        """ nodes that have no outgoing edges """
        return nodes.difference(set(edges.keys()))

    @staticmethod
    def _group_by(lst, func):
        res = {}
        for item in lst:
            k = func(item)
            ADG._add_edge(res, k, item)
        return res

    @staticmethod
    def _bn(name):
        """ Get the block decision name """
        return name.split('-')[0].split(':')[0]

    def _merge_one(self, prev, cur):
        groups = ADG._group_by(cur, ADG._bn)
        if prev:
            self.nodes.add(prev)
            for k in groups.keys():
                ADG._add_edge(self.edges, prev, k)
                self.nodes.add(k)
        return groups

    def _merge(self):
        """ Merge alternatives """
        src = ADG._get_source(self._graph_nodes, self._graph_edges)
        groups = self._merge_one(None, src)
        nds = list(groups.keys())
        i = 0
        while len(nds):
            nd = nds.pop()

            # look up the alternatives, then restore the correct node id
            alts = groups[nd]
            nd = nd.split('-')[0]

            # find the children of all alts of this node and perform merge
            cur = [self._graph_edges[n] for n in alts if n in self._graph_edges]
            cur = [item for sublist in cur for item in sublist]
            # print(nd, set(cur))
            gp = self._merge_one(nd, set(cur))

            # if the child node is already in groups, give it a different id
            for g in gp:
                val = gp[g]
                key = '{}-{}'.format(g, i) if g in groups else g
                i += 1 if g in groups else 0
                del gp[g]
                gp[key] = val

            # update the loop
            groups.update(gp)
            nds.extend(gp.keys())

        # print(self.nodes, self.edges)

    def _prune_recur(self, node, nodes, edges):
        """ Recursive helper for prune. Make sure the graph has no cycles! """
        # leaf node
        if node not in self.edges:
            return [node] if node in self._decs else None

        clean = []
        # recursively prune children
        for nd in self.edges[node]:
            ret = self._prune_recur(nd, nodes, edges)
            if ret:
                clean.extend(ret)
            elif len(self.edges[node]) > 1:  # preserve branches
                clean.append(nd)

        # skip if not decision, else add to edges
        if node in self._decs:
            nodes.update(clean)
            nodes.add(node)
            for nd in clean:
                ADG._add_edge(edges, node, nd)
            return [node]
        else:
            return clean

    def _prune(self):
        """ Remove non-decision nodes """
        edges = {}
        nodes = set()
        src = ADG._get_source(self.nodes, self.edges)
        for s in src:
            self._prune_recur(s, nodes, edges)

        # replace nodes and edges
        self.nodes = nodes
        self.edges = edges

    def _get_linked_vars(self, blocks):
        """ Get linked placeholders """
        bd = set([blocks[b].parameter for b in blocks if blocks[b].parameter])
        res = set()
        for l in self._links:
            bls = [b for b in l if b in bd]
            if len(bls):
                # skip all vars if they are linked with blocks
                res.update(set(l).difference(set(bls)))
            else:
                # otherwise, skip all vars except the first
                res.update(l[1:])
        return res

    def set_graph(self, nodes, edges):
        """ Set code graph """
        self._graph_nodes = nodes
        self._graph_edges = ADG._convert_edges(edges)

    def set_constraints(self, links, proc):
        """ Save the intermediate data from constraint parser """
        self._constraint_proc = proc
        self._links = links

    def create(self, blocks):
        """ Create the ADG """
        # abort if ADG has already been created
        if len(self.nodes):
            return

        # add placeholder vars to the code graph
        decs = []
        for bl in blocks:
            # get the variables associated with a block
            vs = [chunk.variable for chunk in blocks[bl].chunks
                  if chunk.variable != '']
            decs.extend(vs)

            # remove linked vars
            linked = self._get_linked_vars(blocks)
            vs = [v for v in vs if v not in linked]

            # remove duplicates within this block
            tmp = []
            [tmp.append(v) for v in vs if v not in tmp]
            vs = tmp

            # skip variables that have appeared in previous blocks
            # fixme
            gp = ADG._group_by(self._graph_nodes, ADG._bn)
            vs = [v for v in vs if v not in gp
                  or gp[v][0].split('-')[1].split(':')[0] == ADG._bn(bl)]

            # name the placeholders differently as distinct nodes for now
            vs = ['{}-{}'.format(v, bl) for v in vs]
            self._graph_nodes.update(vs)

            # move children of block to the last var
            vs = [bl] + vs
            last = vs[len(vs) - 1]
            if bl in self._graph_edges:
                temp = self._graph_edges[bl]
                self._graph_edges[bl] = []
                self._graph_edges[last] = temp

            # add edges between vars
            for i in range(len(vs) - 1):
                ADG._add_edge(self._graph_edges, vs[i], vs[i + 1])

        # save all decisions, including placeholders and decision blocks
        bd = set([blocks[b].parameter for b in blocks if blocks[b].parameter])
        self._decs = set(decs).union(bd)

        # infer ADG from the graph
        self._merge()
        self._prune()

        # any branch should be a procedural branch
        for s in self.edges:
            t = self.edges[s]
            if len(t) > 1:
                self.proc_edges[s] = t
        # add the procedural deps from constraint
        for proc in self._constraint_proc:
            s = proc.split('-')[0]
            e = proc.split('-')[1]
            ADG._add_edge(self.proc_edges, s, e)

        # todo: remove linked blocks if they don't have procedural branches

    def get_used_decs(self):
        """ Get the decisions that are used in the ADG """
        return [n for n in self.nodes if n in self._decs]

    def output(self):
        """ Output the graph object in server JSON """
        nodes = []
        edges = []

        # nodes
        i = 0
        lookup = {}
        for n in self.nodes:
            nodes.append({"id": i, "name": n})
            lookup[n] = i
            i += 1

        # first add procedural edges
        done = set()
        for s in self.proc_edges:
            ts = self.edges[s]
            for t in ts:
                done.add('{}->{}'.format(s, t))
                edges.append({"source": lookup[s], "target": lookup[t],
                              "type": "procedural"})

        # add order edges, skip those already added
        for s in self.edges:
            ts = self.edges[s]
            for t in ts:
                if '{}->{}'.format(s, t) not in done:
                    edges.append({"source": lookup[s], "target": lookup[t],
                                  "type": "order"})

        return {"graph": {"nodes": nodes, "edges": edges}}



# -*- coding: utf-8 -*-

from dataclasses import dataclass
from .baseparser import BaseParser, Token, ParseError


@dataclass(frozen=True)
class Edge:
    start: str
    end: str


class GraphParser(BaseParser):
    def __init__(self, graph_spec):
        line = '\n'.join(graph_spec)
        super(GraphParser, self).__init__(line)

        self.spec = graph_spec
        self.nodes = set()
        self.edges = set()

    def _prep_err(self, msg):
        return 'At character {} of "{}":\n\t{}'.format(self.col+1, self.spec[self.row], msg)

    def _read_next(self):
        self._read_while(GraphParser._is_whitespace)
        if self._is_end():
            return Token('eof', '')

        ch = self._peek_char()
        if ch == '-':
            return self._read_edge()
        elif GraphParser._is_id_start(ch):
            return self._read_node()
        else:
            raise ParseError(self._prep_err('Cannot handle character "{}"'.format(ch)))

    def _read_edge(self):
        val = self._next_char()
        ch = self._peek_char()
        if ch != '>':
            raise ParseError(self._prep_err('Expected "->", got "-{}"'.format(ch)))
        val += self._next_char()
        return Token('edge', val)

    def _read_node(self):
        nd = self._read_while(GraphParser._is_id)
        return Token('node', nd)

    def parse(self):
        prev_node = None

        while True:
            tk = self._next()

            if tk.type == 'node':
                self.nodes.add(tk.value)
                prev_node = tk.value
            if tk.type == 'edge':
                if not prev_node:
                    raise ParseError(self._prep_err('Cannot find a source node'))
                nx = self._peek()
                if nx.type != 'node':
                    raise ParseError(self._prep_err('Cannot find a target node'))
                self.edges.add(Edge(prev_node, nx.value))
            if tk.type == 'eof':
                break

        return self.nodes, self.edges

    def replace_graph(self, decs):
        """ Replace the block-level decision nodes in the graph with option nodes."""
        # replace nodes
        nds = []
        for nd in self.nodes:
            tmp = decs[nd] if nd in decs else [nd]
            nds.extend(tmp)

        # replace edges
        egs = []
        for eg in self.edges:
            ss = decs[eg.start] if eg.start in decs else [eg.start]
            es = decs[eg.end] if eg.end in decs else [eg.end]
            egs.extend([Edge(s, e) for s in ss for e in es])

        self.nodes = set(nds)
        self.edges = set(egs)
        return self.nodes, self.edges

    def create_default_graph(self, nodes):
        """
        Create the default graph, which is a linear flow of blocks, with the
        same order as they appear in the template script.
        :param nodes: A list of unique blocks.
        :return: nodes and edges
        """
        self.nodes = set(nodes)
        self.edges = set()
        for i in range(len(nodes) - 1):
            self.edges.add(Edge(nodes[i], nodes[i + 1]))
        return self.nodes, self.edges
